@rem = '--*-Perl-*--';
@rem = '
@echo off
if not exist %0 goto next1
cqperl %0 %*
goto endofperl
:next1
if not exist %0.bat goto next2
cqperl %0.bat %*
goto endofperl
:next2
cqperl -S %0 %*
goto endofperl
@rem ';

# msimportauto.bat -- MultiSite Automation
#
# This script automates routine MultiSite tasks, by issuing the 
# appropriate MultiSite commands to synchronize the replicas.
#
# There are many valid approaches to keeping replicas up-to-date using
# either the Multisite shipping server or simple directories as storage 
# areas. When using the shipping server, this program always uses the 
# '-fship' option to sending updates (to tell the shipping server to send
# them immediately). Another approach would be to use 'syncreplica -ship'
# (to stage the updates in the output shipping bay) and then to periodically
# use 'mkorder -fship' to tell the shipping server to actually ship the
# packets. As an alternative to the shipping server, a directory can be 
# specified as a location to store packets.  There are probably other 
# approaches that would work as well...
#
# This program operates in one of two modes:
# 1. Sync now -- The program will receive any pending updates, then
#    send any pending updates, then exit. Use this mode if you want to run
#    the synchronization immediately or if you want to schedule the program's
#    execution via some external scheduler package, such as the Windows
#    "Scheduled Tasks" faclility or the ClearCase scheduler.
# 2. Loop and wait -- The program receives, sends, then sleeps a specified
#    number of seconds, then loops back and receives, sends, and sleeps
#    again. Use this mode if you want the program to essentially act as its
#    own schduler.
#
# The choice of mode is determined by the command-line operands:
#   -debug , -MaxLoops, and -TimeToWait
#
# The default is "Sync now" mode, where the program performs the
# synchronization commands once each and then exits. If 'MaxLoops' is
# specified but TimeToWait is not, then the default is 30 seconds between
# iterations.
#
# Any number of clans/families can be synchronized during the execution of this 
# program.  As a result, multiSite clan information must also be specified by 
# command line arguments. At least 1 clan must be specified from the command line  
#
# The full usage is:
#
# 	msimportauto [-debug NN] [-MaxLoops Num_Loops [-TimeToWait Time_In_Seconds]]
#	[-AndDoExport] -clan clan_1_name clan_1_info
#	[-clan clan_2_name clan_2_info]
#	...
#	[-clan clan_n_name clan_n_info]...
#
# where clan_#_info takes the form (no spaces allowed):
#	admin_username,admin_password;storage_class | directory;
#       family_1,my_site,other_site_1[,other_site_2,]...[,other_site_n][;family_2,my_site,other_site_1...]...
#	[;family_n,my_site,other_site_1[,other_site_2,]...[,other_site_n]]
#
#	my_site is the local site which will be imported into and exported from.  
#	other_site_#'s are the other sites in this clan to be exported to/imported from
#
# An example of usage is:
#
# msimportauto -debug 0 -MaxLoops 2 -TimeToWait 30 -clan TESTCLAN admin,"";
# cq_default;te,SITEA,SITEB,SITEC;te2,SITEA,SITEB -clan TESTCLAN2 admin,"";c:\TESTCLAN2;d2,SITEA,SITEB
# -clan TESTCLAN3 admin,"";cq_default;dt3,SITEA,SITEB -AndDoExport
#
# In this example 3 clans (TESTCLAN, TESTCLAN2, and TESTCLAN3) are synchronized.  Clan
# TESTCLAN consists of 2 families.  Clans TESTCLAN and TESTCLAN3 use the shipping server
# while TESTCLAN2 uses directory c:\TESTCLAN2 to store packets.
#
# This program writes out messages indicating what's going on, and the
# messages go both to STDOUT (normally a command prompt console window) 
# *and* to a 'log' file. The log file is, by default, located in the CQ
# program directory and it has the same name as this program with a suffix
# of ".log" appended. Each execution appends new lines to the end of the file.

use File::Basename;

# Set the debug level:
# 0    = Do real work; don't produce any debugging output
# 1..9 = Show diagnostic info and actually do work too (higher
#        numbers show more granular output)
# 10+  = Show diagnostic info, don't actually do real work
# NOTE: This can also be set from the command-line, using the '-debug NN' command-line option.
$DEBUG = 0;

$MeVersion = "1.0";  # This program's version


############################################################################################################
# Build the commands needed to sync all clans, sites, and families
#
# Returns: ($exportCMDs, %importCMDs) where
# $exportCMDs is a string of SyncExport commands (separated by "\n")
# %importCMDs is a hash with key = clan name and value a string of sync import commands (separated by "\n")
#
############################################################################################################
sub MkMSCommands() {
    my ($exportCMDs) = "";
    my (%importCMDs);


    &MsgOut(2, "MkMSCommands() starting...\n");
    # First, query the registry to get a list of the MultiSite "clans"
    # (schema repo's) we are a member of, and our "site" name within each
    # clan...
    use Win32::TieRegistry;
    $CQDBRegKey = "HKEY_LOCAL_MACHINE\\SOFTWARE\\Rational Software\\ClearQuest\\$CQ_CurrentVersion\\Core\\Databases";
    if (! ($DBSetList = $Registry->{$CQDBRegKey})) {
        &MsgOut(0, "ERROR: Can't read registry key '$CQDBRegKey'\n"
                   . "Unable to determine MultiSite clan membership. Exiting.\n");
        &AllDone(1);
        }
    foreach ( keys (%$DBSetList) ) {
        if (/^CQMS\.(\w+)\.(\w+)\\$/) {
            &MsgOut(3, "Registry indicates clan='$1' site='$2'.\n");
            $MyClanList{"$1 $2"} = 1;
        }
    }

    # For each clan (schema repo) we're a member of, build the commands
    # necessary to receive its updates and to send updates to each of
    # the sites containing any replicas of any of its families
    # (databases)...


    PARSECLANS: foreach $MyClanInfo (sort keys %MyClanList) {
        my($MyClan, $MySite) = split(/ /, $MyClanInfo);
        # Get all the data for this clan; note we can't just look it up by referring
        # to the appropriate hash key, because case-sensitivity seems to get in the
        # way (registry access seems to upcase everything)...
        &MsgOut(5,"Beginning enhanced clan-lookup. MyClan=$MyClan...\n");
        foreach (sort keys %MSClan) {
            if (/^$MyClan$/i) {
                &MsgOut(5,"MyClan='$MyClan'. This entry='$_'... Match!\n");
                # This one matches, possibly without respect to case. Reset 'MyClan' so we
                # can do easier lookups from here on down...
                $MyClan = $_;
                last;
            } else {
                &MsgOut(5,"MyClan='$MyClan'. This entry='$_'... No match.\n");
            }
        }
        &MsgOut(5,"Done enhanced clan-lookup. MyClan=$MyClan...\n");
        if (! $MSClan{$MyClan}) {
            &MsgOut(1, "INTERNAL ERROR: Clan '$MyClan' was not found in internal table 'MSClan'.\n"
                      ."Ignoring this clan...\n");
            delete $MyClanList{$MyClanInfo};
            next;
        }
        my($CQAdminUser, $CQAdminPswd, $StorageClassOrDir, $ClanInfo) = split (" ", $MSClan{$MyClan});
        &MsgOut(3, "MyClan=$MyClan MySite=$MySite CQAdminUser=$CQAdminUser " .
                   "CQAdminPswd=$CQAdminPswd StorageClassOrDir=$StorageClassOrDir ClanInfo=$ClanInfo\n");

        foreach $FamilyInfo (split(";", $ClanInfo)) {
            &MsgOut(3, "FamilyInfo=$FamilyInfo\n");
            my(@OtherSites) = ();
            my($IAmInFamily) = 0;
            if ($FamilyInfo =~ /(.*)@(.*)/) {
                ($Family, $Sites) = ($1, $2);
                &MsgOut(3, "FamilyInfo form #1. Family=$Family, Sites=$Sites.\n");
                # First, see if our site is a member (don't try to synchronize if not!)...
                foreach $Site (split (",", $Sites)) {
                    if ( (uc $Site) eq (uc $MySite)) {
                        $IAmInFamily = 1;
                    } else {
                        push (@OtherSites, $Site);
                    }
                }
            } elsif ($FamilyInfo =~ /(.*):(.*)/) {
                ($Family, $Sites) = ($1, $2);
                &MsgOut(3, "FamilyInfo form #2. Family=$Family, Sites=$Sites.\n");
                foreach $RouteSpec (split (",", $Sites)) {
                    &MsgOut(5, "RouteSpec = '$RouteSpec'.\n");
                    my($S1,$D,$S2) = ($RouteSpec =~ /(\w+)([\<=\>]*)(\w+)/);
                    &MsgOut(5, "S1='$S1' D='$D' S2='$S2'\n");
                    # Is "my" site on the "from" side of this spec?
                    if ($D eq "=>") {
                        # Single-directional route spec...
                        if ( (uc $S1) eq (uc $MySite)) {
                            $IAmInFamily = 1;
                            push (@OtherSites, $S2);
                        }
                    } elsif ($D eq "<=>") {
                        # Bi-directional route spec...
                        if ( (uc $S1) eq (uc $MySite)) {
                            $IAmInFamily = 1;
                            push (@OtherSites, $S2);
                        } elsif ( (uc $S2) eq (uc $MySite)) {
                            $IAmInFamily = 1;
                            push (@OtherSites, $S1);
                        }
                    } else {
                        &MsgOut(1, "INTERNAL ERROR: Cannot parse RouteSpec=$RouteSpec for clan $MyClan.\n"
                                  ."Ignoring this clan...\n");
                        delete $MyClanList{$MyClanInfo};
                        next PARSECLANS;
                    }
                }
            } else {
                &MsgOut(1, "INTERNAL ERROR: Cannot parse FamilyInfo=$FamilyInfo for clan $MyClan.\n"
                          ."Ignoring this clan...\n");
                delete $MyClanList{$MyClanInfo};
                next PARSECLANS;
            }
            &MsgOut(3, "Done route parse. IAmInFamily=$IAmInFamily. MySite=$MySite. OtherSites=@OtherSites.");
            if ($IAmInFamily) {
                &MsgOut(3, "MySite=$MySite. OtherSites=@OtherSites.");
                # Build the "syncreplica -import" command to receive updates
                # from all sites...

                my($Ship) = '-fship';
		
		my $receiveCmd;
		my $sendCmd;

		if (-d $StorageClassOrDir){ 
			$receiveCmd = " $StorageClassOrDir";
			$packetfile = create_packet_file_name($MySite, $StorageClassOrDir, $Family);
			$sendCmd = " -out $packetfile"; 
		}
		else{
			$receiveCmd = " -receive -sclass $StorageClassOrDir";
			$sendCmd = " $Ship -wor $MyWorkDir -sclass $StorageClassOrDir";

		}
		

                $importCMDs{$MyClan} .= ("multiutil syncreplica -imp"
                      . " -u $CQAdminUser"
                      . " -p $CQAdminPswd"
                      . " -clan $MyClan"
                      . " -family $Family"
                      . " -site $MySite"
                      . " $receiveCmd"
                      . "\n"
                      );

                # Build the 'syncreplica -export' command to send updates
                # to all sites...
                ##$Ship = ($MyClan eq 'RAMBU') ? '-ship' : '-fship';
                $exportCMDs .= ("multiutil syncreplica -exp"
                      . " -u $CQAdminUser"
                      . " -p $CQAdminPswd"
                      . " -clan $MyClan"
                      . " -fam $Family"
                      . " -site $MySite"
                      . " $sendCmd"
                      . " @OtherSites"
                      . "\n"
                      );
                 }
        }
    }
    &MsgOut(2, "MkMSCommands() done.\n");
    return ($exportCMDs, %importCMDs);

}

#######################################################################################################
# MkTimeStamp - get a timestamp
#
# returns: a formatted timestamp
#######################################################################################################
sub MkTimeStamp {
    local ($sec,$min,$hr,$mday,$mon,$yr,$wday,$yday,$isdst);
    ($sec,$min,$hr,$mday,$mon,$yr,$wday,$yday,$isdst) = localtime(time);
    # Note: You have to add '1900' to the year and '1' to the month to get
    # human-recognizable results (see localtime() description for details).
    return sprintf ("%04d%02d%02d%02d%02d",$yr+1900,$mon+1,$mday,$hr,$min);
}

#######################################################################################################
# MkMsgHeader - makes a message header
#
# Returns: The message header (which is prepended to all messages)
#
#######################################################################################################
sub MkMsgHeader() {
    return (&MkTimeStamp() . " " . $Me . "[" . $$ . "] ");
}

#######################################################################################################
# MsgOut - outputs a message
#
# $DebugLevel - the debug level of the program
# $Msg - message to output
#
#######################################################################################################
sub MsgOut() {
    local($DebugLevel, $Msg) = @_;
    if ($DebugLevel <= $DEBUG) {
        foreach $ML ((split('\n', $Msg))) {
	    my $M;
	    if ($NoMsgHeader){
		$M = $ML . "\n";
	    }
            else{	
            	$M = &MkMsgHeader() . $ML . "\n";
	    }
            print $M;
            # If we've figured out where the LOG file is, open it, write this
            # line, and close. If we haven't yet figured that out, queue this
            # line so it'll be written eventually...
            if ($MyLogFile ne "") {
                # Yes, we know where the log file is. Open it, send any
                # pending messages, then this message, then clear the
                # pending messages.
                open (LOG, ">>$MyLogFile");
                print LOG $PendingLogMessages . $M;
                $PendingLogMessages = "";
                close LOG;
            } else {
                $PendingLogMessages .= $M;
            }
        }
    }
}

#######################################################################################################
# DOsystem - Run a system command.
#
# args: $CMD = a system command
#
#######################################################################################################
sub DOsystem() {
    my($CMD) = shift;
    &MsgOut(1, "DOSystem($CMD)...\n");
    `$CMD` if ($DEBUG <= 9);
}

########################################################################################################
# AllDone - Cleans up and exits program with return code.
#
# Args: $RC = return code
#       $Msg = message to output
#
########################################################################################################
sub AllDone() {
    my($RC, $Msg) = @_;
    # Clean up the PID file, if we created it...
    if ($RunningPID == $$) {
        &MsgOut(2, "Removing our PID file $MyPIDFile.\n");
        unlink $MyPIDFile;
    } else {
        &MsgOut(2, "Not removing PID file $MyPIDFile (this task didn't create it).\n");
    }
    if ($Msg) {
        &MsgOut(0, "$Msg\n");
    } elsif ($RC == 0) {
        &MsgOut(0, "Exiting normally.\n");
    } else {
        &MsgOut(0, "Exiting with error code=$RC.\n");
    }
    exit $RC;
}

########################################################################################################
# RmDir
#
# Remove a directory including any subordinate files or subdirectories...
#
# Returns: none
########################################################################################################
sub RmDir() {
    my($D) = shift;
    &MsgOut(1, "RmDir($D) starting...\n");
    if (-f $D) {
        &MsgOut(2, "$D is a file... deleting...\n");
        unlink $D if ($DEBUG < 10);
    } elsif (-d $D) {
        &MsgOut(2, "$D is a directory... calling RmDir to remove files in it...\n");
        foreach (<$D\\*>) {
            &RmDir($_);
        }
        &MsgOut(2, "Back from removing files from $D. Removing it...\n");
        rmdir $D if ($DEBUG < 10);
    }
    &MsgOut(1, "RmDir($D) done.\n");
}

#########################################################################################################
# GetRunningPID
#
# Returns: the PID that is (supposed to be) running, from the PIDFile...
# 
#########################################################################################################
sub GetRunningPID() {
    my($MyPIDFile) = shift;
    my($RunningPID);
    if (-f $MyPIDFile) {
        if (! open(PID, "<$MyPIDFile")) {
            &MsgOut(0, "Concurrency problem: Can't open $MyPIDFile for input: $!\n");
            &AllDone(1);
        } else {
            $RunningPID = <PID>;
            close PID;
            chomp $RunningPID;
        }
    }
    &MsgOut(1, "Got RunningPID='$RunningPID' from PID file...\n");
    $RunningPID;
}

########################################################################################################
#Run_Commands - run a list of system commands
#
#args $Command_List = a string - '\n' separated list of commands
#
########################################################################################################
sub Run_Commands($){
    my ($Command_List) = @_;

    # Loop through the list of commands...
    foreach $CMD (split ("\n", $Command_List)) {
        &MsgOut(0, "COMMAND: $CMD\n");
        $Result = &DOsystem($CMD);
        chomp($Result);
        if ($Result eq "") {
            &MsgOut(0, "Command completed. No output.\n");
        } else {
            &MsgOut(0, "Command completed. Output was:\n" . 
                  "$Result\n" .
                  "End of command output\n");
            # Work around defect RAMBU15735 ("Packet file doesn't get deleted") by parsing
            # the command output looking for messages indicating it tried to delete the 
            # packet file. If found, go out and actually delete the file...
            foreach (split ("\n", $Result)) {
                if (/^multiutil: Deleting packet (.*)$/) {
                    $F = $1;
                    &MsgOut(0, "Deleting '$F'...\n");
                    $NDeleted = unlink $F;
                    if ($NDeleted != 1) {
                        &MsgOut(0, "ERROR: Unable to delete $F.\n  Error=$!.\nContinuing...\n");
                    }
                }
            }
        }

        # Work around the problem where sometimes these commands don't get rid of
        # the workdir. If the workdir still exists, delete it...
        if (-d $MyWorkDir) {
            &MsgOut(0, "$MyWorkDir not cleaned up by command. Deleting...\n");
            &RmDir($MyWorkDir);
        }
    }

}


#######################################################################################################
#get_file_import_directory($)
#
#args $storage_class
#
#returns - path to the "incoming" bin for the shipping server
#
#######################################################################################################
sub get_file_import_directory($){
	#return the "incoming" bin for the shipping server
	my ($storage_class) = @_;

	use Win32::TieRegistry;
	$CQDBRegKey = "HKEY_LOCAL_MACHINE\\SOFTWARE\\Atria\\ClearCase\\CurrentVersion\\MultiSite\\StorageClass\\$storage_class";

	if (! ($mskey = $Registry->{$CQDBRegKey})) {
        &MsgOut(0, "ERROR: Can't read registry key '$CQDBRegKey'\n"
                   . "Unable to determine shipping server incoming bay.\n");
        &AllDone(1);
        }

	$valueString= $mskey->GetValue("StorageBay");
	$valueString .= "\\incoming";

	return $valueString;
}


#################################################################################################
#directories_have_changed
#
#Given 2 directories see if the contents have changed
#
#args:
#	$directory_ref - reference to sorted array list of directory files which may have changed
#	$previous_directory_ref - reference to sorted array list of original files
#
#
#returns: 1 = changed; 0 the same
#
#################################################################################################
sub directories_have_changed($$){
	my $directory_ref;
	my $previous_directory_ref;

	($directory_ref, $previous_directory_ref) = @_;	


	$size_prev_dir = @$previous_directory_ref;
	$size_new_dir = @$directory_ref;
	if ($size_prev_dir != $size_new_dir){
		&MsgOut(1, "Directories have changed!\n");
		return 1;
	}

	for ($filecount = 0; $filecount < $size_prev_dir; $filecount++){
		if ($previous_directory_ref->[$filecount] ne $directory_ref->[$filecount]){
			&MsgOut(1, "Directories have changed!\n");
			return 1;
		}
	}

	&MsgOut(1, "Directories are the same!\n");
	return 0;

}






#################################################################################################
#get_file_list_from_directory
#
#Given a directory, return an array containing each file in the directory
#
#arg: $directory - the directory name (path)
#returns: @filelist - a sorted array of filenames
#
#################################################################################################
sub get_file_list_from_directory($){
	my $directory;

	($directory) = @_;

	if (! -d $directory){
		die "$directory is not a directory!";
	}
	
	my @dir_ents;

    	opendir (CQMSDIRECTORY, $directory) or die "couldnt open directory!: $!";
    	@dir_ents = readdir CQMSDIRECTORY;
    	closedir(CQMSDIRECTORY);

    	my (@filelist) = grep { -f $directory . $FS . $_ } @dir_ents;

    	return sort @filelist;
}

##########################################################################################################
#sync_import_routine:
#
#Does the following:
#foreach clan
#	filelist = get the list of files in the directory to sync import on
#	while filelist is not empty
#		for each family in the current clan issue syncreplica imports
#		end foreach family 
#		check whether the contents of the directory have changed
#		if directory has changed or is now empty set the filelist = newlist
#		elseif the directory has not changed - last (break the while loop + process the next clan)
#	end while	
#end foreach clan
#
#
#args: %import_commandlist - a hash with key of clan, value = string of sync imp commands for the clan
#
##########################################################################################################
sub sync_import_routine(%){

	my (%import_commandlist) = @_;	

	#################################################################################################
	#for each clan try to do a sync import on all families, repetitively if necessary
	foreach $MyClanInfo (sort keys %MSClan){

		&MsgOut(2, "clan being imported is ... $MyClanInfo cmds =:\n $import_commandlist{$MyClanInfo}");

		$imp_cmds_for_clan = $import_commandlist{$MyClanInfo};

		##########################################################################################
		#determine what the directory to process is - by parsing that info out of the command list
		#- directory or storage class (or packet) is the last argument of any sync import command
		##########################################################################################

		@impcmds = split ("\n", $imp_cmds_for_clan); 		#get one sync import cmd
		@elements_from_one_cmd = split(" ", $impcmds[0]);	#break the command into its elements
		$array_sz = scalar(@elements_from_one_cmd);		
		$directory_or_sclass = @elements_from_one_cmd[array_sz - 1]; #the last element in the command

		if (-d $directory_or_sclass){
			$directory_to_process = $directory_or_sclass;
		}
		elsif (-f $directory_or_sclass){
			###tbd will we support this???  - just importing a single file
			#get the parent directory of $directory_or_sclass
		}
		else{
			#assume that the last arg is a storage class
			$directory_to_process = get_file_import_directory($directory_or_sclass);
		}
	

		#get current list of xml files in the directory we're importing from	
		@MyOrigXmlFiles = get_file_list_from_directory($directory_to_process);
		$ref_MyOrigXmlFiles = \@MyOrigXmlFiles;

		#while the directory being imported from is not empty attempt to sync import
		while (scalar(@$ref_MyOrigXmlFiles) > 0){

			#run the import commands
 			Run_Commands($imp_cmds_for_clan);	

			#Check if the list has changed:
			#Get the list of the files in the directory now.
			#If it has changed and files remain in the directory then we need to attempt another import.
			#if it hasnt changed and files remain in the directory we've got a problem so break out of this
			@MyXmlFiles = get_file_list_from_directory($directory_to_process);

			if (scalar(@MyXmlFiles) > 0){
				$ref_MyXmlFiles = \@MyXmlFiles;		

				if (! directories_have_changed($ref_MyXmlFiles, $ref_MyOrigXmlFiles)){
					&MsgOut(0, "Clan: $MyClanInfo - Cannot import any more in directory $directory_to_process" .
						".  Directories havent changed following last sync import attempt!\n");
					last; 	#quit trying to sync import this clan
				}
				else{
					#directories have changed and there are remaining files in the directory
					#we need to retry sync import - re-process the directory
					$ref_MyOrigXmlFiles = \@MyXmlFiles;
				}
			}	
			else{
				#Everything has been processed - we're done with this clan
#				&MsgOut(0, "Clan: $MyClanInfo - Everything in $directory_to_process has now been imported\n");
				$ref_MyOrigXmlFiles = \@MyXmlFiles;
			}	

		}
	
		&MsgOut(0, "Clan: $MyClanInfo - Nothing else needs to be imported in: $directory_to_process \n");
		next; #go to the next clan and try to sync import
	}

	return 0;

}

#########################################################################################################
#create_packet_file_name
#
#args: site
#returns: a unique name for packet file (ex: sync_SITEA_16-July-02_10-08-19.xml)
#########################################################################################################
sub create_packet_file_name($$){
	my($pkt_site, $pkt_path, $pkt_family) = @_;
	 
	($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime;
	$year = sprintf("%02d", $year % 100);
	$mon++;
	$file_ext = ".xml";
	
	$filename = "$pkt_path$FS" . "sync_" . $pkt_site . "_$pkt_family" . "_$mday" . "-$mon" . "-$year" . "_$hour" . "-$min" . "-$sec" . "$file_ext";
	&MsgOut(3,"filename: $filename\n");

	return $filename;
}

###############################################################################################################
#Process any and all command line arguments.
#
#args: @ARGV
#
#
#usage:	msimportauto [-debug NN] [-MaxLoops Num_Loops [-TimeToWait Time_In_Seconds]] [-AndDoExport] 
#	-clan clan_1_name clan_1_info
#	[-clan clan_2_name clan_2_info]
#	...
#	[-clan clan_n_name clan_n_info]...
#
# where clan_#_info takes the form (no spaces allowed):
#	admin_username,admin_password;storage_class | directory;
#       family_1,my_site,other_site_1[,other_site_2,]...[,other_site_n][;family_2,my_site,other_site_1...]...
#	[;family_n,my_site,other_site_1[,other_site_2,]...[,other_site_n]]
#
#	my_site is the local site which will be imported into and exported from.  
#	other_site_#'s are the other sites in this clan to be exported to/imported from
#
###############################################################################################################
sub process_command_args(@){

	my(@cmd_line_args) = @_;

	$usage_text = "usage: \nmsimportauto [-debug NN] [-MaxLoops Num_Loops [-TimeToWait Time_In_Seconds]]\n" .
			"[-AndDoExport] -clan clan_1_name clan_1_info\n" .
			"[-clan clan_2_name clan_2_info]...[-clan clan_n_name clan_n_info]\n\n" .
			"where clan_#_info takes the form (no spaces):\n" .
			"admin_username,admin_password\;storage_class | directory;\n" . 
			"family_1,my_site,other_site_1[,other_site_2]...[,other_site_n]\;\n" . 
			"[family_2,my_site,other_site_1...];...[family_n,my_site,other_site_1...];\n\n" .
			"my_site is the local site which will be imported into and exported from.\n" .  
			"other_site_#'s are the other sites in this clan to be exported to/imported from\n";

	my $got_clan_info = 0;
	my $arg_is_ok;

	$size_arg_list = $#cmd_line_args;

	for ($arg_num = 0;$arg_num <= $size_arg_list;$arg_num++){
		$cmd_line_arg = @cmd_line_args[$arg_num];
		
		$arg_is_ok = 0;

		SWITCH: {
			if ($cmd_line_arg eq "-NoMsgHeader"){
				$arg_is_ok = 1;
				$NoMsgHeader = 1;
				last SWITCH;
			}
			if ($cmd_line_arg eq "-debug"){
				$arg_is_ok = 1;
				$arg_num++;
				$DEBUG = $cmd_line_args[$arg_num];
				last SWITCH;
			}
			if ($cmd_line_arg eq "-MaxLoops"){
				$arg_is_ok = 1;
				$arg_num++;
				$MaxLoops = $cmd_line_args[$arg_num];
				last SWITCH;
			}
			if ($cmd_line_arg eq "-TimeToWait"){
				$arg_is_ok = 1;
				$arg_num++;
				$TimeToWait = $cmd_line_args[$arg_num];
				last SWITCH;
			}
			if($cmd_line_arg eq "-AndDoExport"){
				$arg_is_ok = 1;
				$AndDoExport = 1;
			}
			if ($cmd_line_arg eq "-clan"){
				$arg_is_ok = 1;
				$got_clan_info = 1;

				$arg_num++;
				$clan_name = $cmd_line_args[$arg_num];
				$arg_num++;
				$full_clan_info = $cmd_line_args[$arg_num];
				
				$clan_info = "";
				@clan_info_elements = split("\;", $full_clan_info);
				$size_CI_elements = $#clan_info_elements;

				if ($size_CI_elements < 2){
					&MsgOut(0, "Incorrect Clan Info Detected:\n");
					&MsgOut(0, "usage_text\n");
					&AllDone(1);	
				}
					
   				for($element_num = 0; $element_num <= $size_CI_elements; $element_num++){
					$clan_info_element = @clan_info_elements[$element_num];
					if ($element_num == 0){
						#get the cq admin username and password
						@cq_login_info = split(",", $clan_info_element);
						$num_creds = scalar(@cq_login_info);
						
						if ($num_creds == 1){
							$admin = $cq_login_info[0];
							$pwd = "\"\"";
						}
						elsif ($num_creds == 2){
							$admin = $cq_login_info[0];
							$pwd = $cq_login_info[1];
						}
						else{
							&MsgOut(0, $usage_text . "\n");
							&AllDone(1);
						}
						$clan_info .= "$admin $pwd ";
						
					}
					elsif ($element_num == 1){
						#get the storage class or working directory
						$storage_info = $clan_info_element;
						$clan_info .= "$storage_info ";
					}
					elsif ($element_num >= 2){
						#get the families and sites
						@family_site_info = split(",", $clan_info_element);
						@family_site_info +0;
						$array_size = $#family_site_info;
						for ($i = 0; $i <= $array_size; $i++){
							if ($i == 0){
								$family = @family_site_info[i];
								$clan_info .= "$family:";
							}
							else {
								$site = @family_site_info[$i];
								if ($i == 1){
									$clan_info .= "$site";
									$site_from = $site;	
								}
								elsif($i == 2){
									$clan_info .= "=>$site";
								}
								else{
									#an additional "site_to"
									$clan_info .= ";$family:" . "$site_from" . "=>$site";
								}


							}
						}
						if ($element_num < $size_CI_elements){
							$clan_info .= "\;";
						}
					}

				} 

				&MsgOut(0, "clan_info parsed from cmd line: = $clan_info\n");

				#construct clan info for this clan
				$MSClan{$clan_name} = $clan_info;

				last SWITCH;
			}

			if (! $arg_is_ok){
				&MsgOut(0, "Bad argument detected: $cmd_line_arg\n");
				&MsgOut(0, "$usage_text\n");
				&AllDone(1);	
			}

		}	

	}

	if (! $got_clan_info){
		&MsgOut(0, "Need at least 1 clan to process.\n");
		&MsgOut(0, "$usage_text\n");
		&AllDone(1);
	}
	
	#the defaults
	$MaxLoops = 1 if (! $MaxLoops);
	$TimeToWait = 30 if (! $TimeToWait);
	

}


#########################################################################################################
# Main program...
#########################################################################################################
# Set platform-specific stuff...
if ($ENV{'OS'} =~ /Windows/) {
	# Windows...
	$FS = "\\";    # Filepath separator...
} else {
	# UNIX...
	$FS = '/';     # Filepath separator...
}

# Get our program name...
$Me = basename($0);

#global info we need to get from the command line:
$DEBUG;
$MaxLoops;
$TimeToWait;
$AndDoExport;
$NoMsgHeader;
%MSClan;

process_command_args(@ARGV);

# Print 'greeting' info...
&MsgOut(0,"$Me: MultiSite Automation Script version $MeVersion starting.\n");
if ($MaxLoops > 1) {
    &MsgOut(0,"Program will loop $MaxLoops times, waiting $TimeToWait seconds between iterations.\n");
}
&MsgOut(1,"DEBUG Level = $DEBUG\n");

# Find out where CQ is installed...
$CQ_CurrentVersion = $Registry->{"HKEY_LOCAL_MACHINE\\Software\\Rational Software\\ClearQuest\\CurrentVersion"};
$CQPath = $Registry->{"HKEY_LOCAL_MACHINE\\Software\\Rational Software\\ClearQuest\\$CQ_CurrentVersion\\Install\\TARGETDIR"};
$MUEXE = $CQPath."\\MultiUtil.exe";
if (! -f $MUEXE) {
    &MsgOut(0, "ERROR: Unable to find ClearQuest directory. Quitting...\n");
    &AllDone(1);
}
&MsgOut(1, "CQ Version  = $CQ_CurrentVersion\n");
&MsgOut(1, "CQ Path     = $CQPath\n");
# Slap the CQ directory onto the front of the path so it's easier to find MultiUtil.exe...
$ENV{'PATH'} = "$CQPath" . ";" . $ENV{'PATH'};

# Build the pathname of the LOG file...
$MyLogFile = $Me;
$MyLogFile =~ s/\.\w+$/\.log/;
$MyLogFile = $CQPath . $FS . $MyLogFile;
&MsgOut(1, "Log file    = $MyLogFile\n");

# Build the pathname of the PID file...
$MyPIDFile = $Me;
$MyPIDFile =~ s/\.\w+$/\.pid/;
$MyPIDFile = $CQPath . $FS . $MyPIDFile;
&MsgOut(1, "PID file    = $MyPIDFile\n");

# Temporary work directory...
$MyWorkDir = $Me;
$MyWorkDir =~ s/\.\w+$/\.TMP/;
$MyWorkDir = $ENV{'TEMP'} . $FS . $MyWorkDir;
&MsgOut(1, "Work dir    = $MyWorkDir\n");

# See if there's already an instance of ourselves running...
&MsgOut(5, "Concurrency check beginning. MyPIDFile='$MyPIDFile'...\n");
$RunningPID = &GetRunningPID($MyPIDFile);
if ($RunningPID) {
    # The PIDFile indicates that another instances of our program is already running.
    # Use 'kill 0' to see if the indicated process actually exists.
    # NOTE: I have experimentally proven that this doesn't work on Perl releases lower
    #       than 5.6 (the target process gets killed, which isn't supposed to happen).
    #       Furthermore, according to the ActivePerl FAQ, signals aren't supported on
    #       Win32 due to limitations in the underlying Windoze API. But it seems to
    #       work so I'm leaving this in here...

    if ($] < 5.00) {
	
        &MsgOut(0, "PROBLEM: $Me [PID=$RunningPID] may already be running, but this program can't\n"
                  ."         (safely) check its status (Perl version is $] -- need 5.006 or above).\n"
                  ."         Assuming the other instance of $Me [PID=$RunningPID] is alive... Exiting.\n"
               );
        &AllDone(1);
    } else {
        # OK to check if process exists by sending it 'kill 0'...
        if (kill 0 => $RunningPID) {
            # Process accepted the signal, so it's probably real (either that or
            # there's another unrelated process that just happens to have picked up
            # the same PID... unlikely, but possible -- and impossible for this code
            # to distinguish)...
            &AllDone(0, "$Me [PID=$RunningPID] is already running. Exiting"); # This is considered normal!
        } else {
            # Process didn't accept the signal; probably doesn't exist.
            if (unlink $MyPIDFile) {
                &MsgOut(1, "PROBLEM (fixed): Running PID $RunningPID not found. "
                          ."Obsolete PID file cleaned up successfully.\n");
                undef $RunningPID;
            } else {
                &MsgOut(0, "PROBLEM: Running PID $RunningPID not found. "
                          ."Obsolete PID file $MyPIDFile could not be removed.\n");
                &AllDone(1);
            }
        }
    }
}
# If we reach this point, there's no other running instance of ourself...
# Write our PID to the PIDFile...
open(PID, ">$MyPIDFile") or die "Can't open $MyPIDFile for output: $!";
print PID "$$";
close PID;
&MsgOut(1, "Wrote my PID ($$) to PID file...\n");
# Reread the PID file to ensure that our PID (which we just wrote there) is still
# there; this escapes a possible race condition between two instances of this
# program both writing the PID file at the same time...
$RunningPID = &GetRunningPID($MyPIDFile);
if ($RunningPID != $$) {
    &MsgOut(0, "PROBLEM: Concurrency-protection PID file race condition detected. "
              ."PID written was $$. PID found was $RunningPID. Exiting.\n");
    &AllDone(1);
} else {
    &MsgOut(1, "Concurrency-check OK; Our PID ($RunningPID) written to PIDFile. Continuing...\n");
}

# Gather info about the clans we're a member of, and using that information build the
# commands that will be used to actually synchronize us with all the sites we're 
# connected to...


my ( $sync_export_CMDS, %sync_import_CMDS ) = MkMSCommands();


&MsgOut(1, "Commands to synch us with other sites = '$MS_Cmds'\n");

# Loop doing work until told to stop...
for ($i = 1; $i <= $MaxLoops; $i++) {
    &MsgOut(0, "Loop iteration $i of $MaxLoops...\n") if ($MaxLoops > 1);
	
    #try running the sync imports - multiple times if necessary for each clan
    sync_import_routine(%sync_import_CMDS);
 
    if ($AndDoExport){
    	#run all of the sync exports
    	Run_Commands($sync_export_CMDS);
    }

    # Sleep the prescribed number of seconds before looping back...
    if ($i < $MaxLoops) {
        &MsgOut(0, "Sleeping $TimeToWait seconds...\n");
        sleep $TimeToWait;
    }
}
&AllDone(0);

__END__
:endofperl
